'' Basic WF system
''
'' Author: MSrobots
'' Version: 1
'' Date: 2018-09-07
''
'' Copyright (c) 2018 Michael Sommer
'' See end of file for terms of use.
''
CON

  'Pin assignments
  VOL = 20       'volume level control pin
  SND = 21       'audio/sound output pin  
  SD_DO = 4      'SD Card DO/MISO
  SD_CK = 5      'SD Card Clock (SCLK)
  SD_DI = 9      'SD Card DI/MOSI  
  SD_CS = 15     'SD Card Chip Select, no pull-up resistor
  'data and clock pin assignments, respectively, in QuadKeyboard driver (in two places)
  'kbd0: 7, 8; kbd1: 12, 13; kbd2: 16, 17; kbd3: 23, 24  

  #0,  NORTH, WEST, SOUTH, EAST, NWSE     '4 scr nums (0..3), counterclockwise from north, plus all 4 screens (4)
  #10, QCN, QCW, QCS, QCE, QCC            '4 QC data for drawing quad chars: N, W, S, E and Colon
                                      
  #8,  BS, TAB, LF, VT, FF, CR, ESC = 27

  #0, MOUNT, UNMOUNT

  'foregnd-backgnd color pairs; avoid KK ($00) = Blk-on-Blk (string delimiter); don't use WR wht-on-red (reserved word)
  KK = $00, KR = $01, KG = $02, KB = $03, KY = $04, KP = $05, KA = $06, KW = $07 'blk foreground on some background (Note 1)
  RK = $10, RR = $11, RG = $12, RB = $13, RY = $14, RP = $15, RA = $16, RW = $17 'red foreground on some background
  GK = $20, GR = $21, GG = $22, GB = $23, GY = $24, GP = $25, GA = $26, GW = $27 'grn foreground on some backgroung  
  BK = $30, BR = $31, BG = $32, BB = $33, BY = $34, BP = $35, BA = $36, BW = $37 'blu foreground on some backgroung 
  YK = $40, YR = $41, YG = $42, YB = $43, YY = $44, YP = $45, YA = $46, YW = $47 'yel foreground on some backgroung
  PK = $50, PR = $51, PG = $52, PB = $53, PY = $54, PP = $55, PA = $56, PW = $57 'pur foreground on some backgroung     
  AK = $60, AR = $61, AG = $62, AB = $63, AY = $64, AP = $65, AA = $66, AW = $67 'aqu foreground on some backgroung  
  WK = $70, WS = $71, WG = $72, WB = $73, WY = $74, WP = $75, WA = $76, WW = $77 'wht foreground on some backgroung (Note 2)
  'Note 1: don't use KK ($00) = Blk-on-Blk in strings since 0 is the string delimiter value (for terminating strings) 
  'Note 2: for white-on-red, use white-on-scarlet (the scarlet letter) = WS because WR is a reserved word in PASM 
  'byte    %%0000, %%0010, %%0020, %%1000  ' 0: Black,      1: Red,        2: Green,      3: Blue  'Quaternary form
  'byte    %%0030, %%1010, %%1020, %%1030  ' 4: Yellow,     5: Purple,     6: Aqua        7: White 'Quaternary form

OBJ
  vga[4] : "vga"                                        'VGA interface to VGA driver
  sdc    : "DS1307_SD-MMC_FATEngine.spin"               'SD card driver (FAT32), can also reload RAM from SD .bin file contains rtc driver   
  kbd    : "QuadKeyboard"                               'Quad keyboard driver (JRG)
VAR
  long vgaLink[4]               'vga linkadresses 
  long volume                   'speaker volume 0..20 (using long to avoid sign extend)   

  long rv                       'generic return value
  long sdErrMsg                 'SD card error string
  long sdErrNum                 'SD card error number
  byte sdProblem                'problem starting SD driver or mounting card; true/false
  byte rtcFound                 'boolean for reporting whether clock module was found
    
DAT '                          ###   NO MAIN LOOP   ###
PUB version
RETURN 1

DAT '                          ### INITIALIZATION ###
PUB initSystem | s, i, vpins, lock

  'config audio pins
  dira[SND]~~                   'audio output pin
  dira[VOL]~~                   'volume control pin
  volume := 10                  'initial volume level  

  'set speaker volume
  ctra :=  %00110_000 << 23 + 1 << 9 + VOL   'counter mode = duty cycle for d/a using rc circuit (B pin is unused)
  frqa := $2000_0000                         'Vout = 3.3 * (frqa/2^32) = 3.3 * 0.5 = 1.65V max when frqa = $8000_0000 (can omit this line) 

  'configure video pins
  lock := ((cnt >> 16 + 2) | 1) << 16   
  repeat s from 0 to 3 
    vga[s].init(lock|vconfig[s])                'initialize vid drv with vgrp/vpins
    vgaLink[s] := vga[s].getscrnadr        

  repeat s from 0 to 3
    'vga[s].setn(1, @pal0)                       'switch to palette 0   
    vga[s].str(string($1B, "s"))                'page mode
    'vga[s].setn(1, constant(NEGX|$8000))        'use the ROM font for chars 0..31
    vga[s].setn(1, NEGX|@user_font)             'use the custom font for chars 0..31
    waitf(4) 'let voltage to screens stablize?

  'start quad keyboard driver
  kbd.start

  waitcnt(clkfreq/4+cnt)          'just in case
  startSDCard                     'start SDC driver and RTC driver
  waitcnt(clkfreq/4+cnt)          'just in case  
  mountSDCard                     'mount card, if present
    
  repeat i from 0 to 3            'defaults, etc.
    kbd.clearKeyBuffer(i)         'empty kbd buffers (probably not strictly necessary)

  checkForRTCModule
    
DAT '                           INPUT FUNCTIONS FOR KEYBOARDS
PUB getKey(s)
RETURN kbd.getKey(s)
 
DAT '                           OUTPUT FUNCTIONS FOR SCREENS
PUB getscrnadr(s)
RETURN vgaLink[s]

PUB getscrnsize(s)
RETURN vga#bcnt_raw

PUB peekChar(s, x, y) | s1, x1, y1
  s1 := 0 #> s <# 3
  x1 := 0 #> x <# 39
  y1 := 0 #> y <# 14  
RETURN byte[vgaLink[s1]+y1*40+x1]

PUB peekColor(s, x, y) | s1, x1, y1
  s1 := 0 #> s <# 3
  x1 := 0 #> x <# 39
  y1 := 0 #> y <# 14  
RETURN byte[vgaLink[s1]+vga#bcnt_raw+y1*40+x1]

PUB pokeChar(s, x, y, value) | s1, x1, y1
  s1 := 0 #> s <# 3
  x1 := 0 #> x <# 39
  y1 := 0 #> y <# 14  
  byte[vgaLink[s1]+y1*40+x1] := value 

PUB pokeColor(s, x, y, value) | s1, x1, y1
  s1 := 0 #> s <# 3
  x1 := 0 #> x <# 39
  y1 := 0 #> y <# 14  
  byte[vgaLink[s1]+vga#bcnt_raw+y1*40+x1] := value

PUB storeScreen(s, filePathName, scrColor) | count
  if scrColor==-2
    count :=  vga#bcnt_raw + vga#bcnt_raw
  else
    count :=  vga#bcnt_raw
  \sdc.openFile(filePathName, "W")
  \sdc.writeData(vgaLink[s], count)
  'if scrColor> -1
  ' writing Text with fixed Color? txt to screen with color?  
  \sdc.closeFile
     
PUB restoreScreen(s, filePathName, scrColor) | count
  if scrColor==-2
    count :=  vga#bcnt_raw + vga#bcnt_raw
  else
    count :=  vga#bcnt_raw
  \sdc.openFile(filePathName, "R")
  \sdc.readData(vgaLink[s], count)
  if scrColor> -1
    bytefill(vgaLink[s] + count, scrColor, count)
  \sdc.closeFile

PUB blitMap(s, filePathName, mapW, mapH, mapX, mapY, scrW, scrH, scrX, scrY, scrColor, borderChar, borderColor)  | ascr, count, spos, fpos, flen, mX, mY, sX, sY, r, mpX, mpY, scX, scY
  if mapW>0 and mapH>0 and scrW>0 and scrW < 41 and scrH>0 and scrH < 16 
    \sdc.openFile(filePathName, "R")     

    count := vga#bcnt_raw ' size of screen buffer
    ascr := vgaLink[s]
    mpX := -(scrW-1) #> mapX <# mapW-1
    mpY := -(scrH-1) #> mapY <# mapH-1
    scX := 0 #> scrX <# (40-scrW)
    scY := 0 #> scrY <# (15-scrH)
        
    if scrColor==-2
      r := 2
    else
      r := 1      
    repeat r
      sY   := scY
      mY   := mpY
      repeat scrH
        sX := scX
        mX := mpX
        spos := ascr + sX + sY * 40
        if (mY < mapH) and (mY > -1)
          if (mapW-mX)<scrW
            flen := mapW-mX
            if ascr==vgaLink[s]
              if borderChar>-1    
                bytefill(spos+flen, borderChar, scrW-flen) 
              if borderColor>-1    
                bytefill(spos+count+flen, borderColor, scrW-flen)    
          else
            flen := scrW
          if mX<0
            if ascr==vgaLink[s]
              if borderChar>-1
                bytefill(spos, borderChar, -mX) 
              if borderColor>-1
                bytefill(spos+count, borderColor, -mX)    
            fpos := mY * mapW
            flen += mX           ' mX is negative thus + and minus swapped
            spos -= mX
          else
            fpos := mY * mapW + mX            
          if ascr<>vgaLink[s]
            fpos += (mapW*mapH)
          \sdc.fileSeek(fpos)
          \sdc.readData(spos, flen)
          if ascr==vgaLink[s]
            if scrColor>-1
              bytefill(spos+count, scrColor, flen)          
        else
          if ascr==vgaLink[s]
            if borderChar>-1
              bytefill(spos, borderChar, scrW) 
            if borderColor>-1
              bytefill(spos+count, borderColor, scrW)    
        mY++
        sY++
      ascr += count
    \sdc.closeFile

PUB setXYC(s, c, r, k)  'set cursor pos and color
  's = screen, c = column, r = row, k = color
  vga[s].out($A)
  vga[s].out(c)
  vga[s].out($B)
  vga[s].out(r)   
  vga[s].out($1B)         
  vga[s].out("c")
  vga[s].out(k)

PUB putChar(s, c, r, k, ch)  'put (not out) character
  's=screen, r=row, c=column, k=color, ch=character 
  setXYC(s, c, r, k)
  vga[s].putc(ch)

PUB putChar4(c, r, k, ch) | s  'put (not out) character to all screens
  's=screen, r=row, c=column, k=color, ch=character
  repeat s from 0 to 3
    setXYC(s, c, r, k)
    vga[s].putc(ch)
  
PUB write(s, c, r, k, str) 'output string to screen s at r, c in color k
  's=screen, c=column, r=row, k=color, str=string to write  
  vga[s].str2(c, r, k, str)

PUB write2(s, c, r, k, addr, n) 'output string to screen s at r, c in color k
  's=screen, c=column, r=row, k=color, str=string to write  
  vga[s].str3(c, r, k, addr, n)

PUB clear(s, k) | i, beg, end  'clear screen to solid background color
's = screen (0..3 or 4 for all screens), k = fore/background color
  beg := end := s
  if s => 4
    beg := NORTH
    end := EAST

  repeat i from beg to end
    vga[i].clear(k)

PUB repChar(s, c, r, k, ch, rep)  'output a single character severals times in a row
'display character ch from column c to column c+rep-1 (rep = repeat) on row r in color k
  setXYC(s, c, r, k)   
  repeat rep
    vga[s].putc(ch)

PUB showNN(s, c, r, k, val) | i, beg, end,d0,d1  'show a two-digit number on one or four screens
'draw two-digit number on one or all four screens
's=screen or use 4 for all, c=column, r=row, k=color, val=value to display
  beg := end := s
  if s == 4
    beg := 0
    end := 3

  d1 := val / 10
  d0 := val // 10
  repeat i from beg to end   
    putChar(i, c, r, k, d1+48)
    putChar(i, c+1, r, k, d0+48)

PUB showHeaderBar(s, k1, k2)
'show header bar on screen s in color k1 at top and middle and in k2 at bottom of header 
  repChar(s, 0, 0, k1, " ", 80)
  repChar(s, 0, 2, k2,  13, 40)  
    
DAT '                              ### SOUND/AUDIO ###
PUB tone(d, f)
'duration (number of wave cycles and relative frequency

  repeat d
    outa[SND]~                     'one wave cycle
    waitcnt(clkfreq/f+cnt)
    outa[SND]~~
    waitcnt(clkfreq/f+cnt) 

PUB beep
  tone(10, 1000)    'short beep, good for debugging
       
DAT                           'Real Time Clock (RTC) Functions
PUB clockDOW
RETURN sdc.clockDOW

PUB clockMTH
RETURN sdc.clockMTH

PUB clockMode
RETURN sdc.clockMode 

PUB clockAMPM
RETURN sdc.clockAMPM

PUB clockSecond
RETURN sdc.clockSecond

PUB clockMinute
RETURN sdc.clockMinute 

PUB clockHour
RETURN sdc.clockHour

PUB clockDay
RETURN sdc.clockDay - 1 'get day of week and zero-base it (SUN..SAT = 0..6)

PUB clockDate
RETURN sdc.clockDate  'range: 1..31 

PUB clockMonth
RETURN sdc.clockMonth - 1 'get month and zero-base it (JAN..DEC = 0..11)

PUB clockYear
RETURN sdc.clockYear  

PUB clockSquareWaveOut(frequency, state) 
RETURN sdc.clockSquareWaveOut(frequency, state)

PUB readClock
RETURN sdc.readClock 'update the clock's output register

PUB setClock(ds_mode, ds_ampm, second, minute, hour, day, date, month, year)
RETURN sdc.setClock(ds_mode, ds_ampm, second, minute, hour, day, date, month, year)

DAT '                                ### SD CARD ### 
PUB sdcall(funct)
  case funct
    0: sdc.mountPartition(0) ' Mount the default partition 0. Can be 0 - 3.
    1: sdc.unmountPartition

PUB startSDCard
  'start FAT driver for SD Card
  'do, clk, di, cs, wp, cd, rtcda, rtcclk, rtcbuslck/i2clock  
  rv := sdc.FATEngineStart(SD_DO, SD_CK, SD_DI, SD_CS, -1, -1, 29, 28, -1)
  sdProblem := !rv
  if !rv
    sdErrMsg := string("Driver could not start")
    rtcFound := false
  else  
    rtcFound := true
    
PUB restartSDC
'restart SD card driver and mount the first partition of the SD card
  \sdc.unmountPartition      'unmount the SD card (if mounted)
  sdc.FATEngineStop          'stop the SD card driver (if running)
  waitcnt(clkfreq/20+cnt)    'give SD card time, just in case
  startSDCard                'try starting the driver again
  mountSDCard                'and try mounting the SD card (if any)
    
PUB mountSDCard
  'mount sd card
  sdProblem := false
  sdErrMsg := \sdcall(MOUNT) 'Returns the address of the error string or null.
  sdErrNum := sdc.partitionError 'Returns the error number or zero.
  if sdErrNum > 0
    sdProblem := true

PUB stopSDCard
  sdErrMsg := \sdc.closeFile 'close the open file, if there is one open
  sdErrMsg := \sdcall(UNMOUNT)
  \sdc.FATEngineStop 

PUB bootPartition(pathfile)
  \sdc.bootPartition(pathfile)

PUB fileSize(pathfile)
  \sdc.openFile(pathfile, "R")
  RESULT := \sdc.fileSize     
  \sdc.closeFile

PUB pokeFile(pathfile, position, value, backgound, offset) | size, offsetpos
  sdByte := 0
  offsetpos := offset
  if offsetpos<0
    offsetpos:= -offset
  \sdc.openFile(pathfile, "W")
  size := \sdc.fileSize
  
  \sdc.fileSeek(size-1)
  \sdc.readData(@sdByte, 1)

  if offset>-1
    if position==size-1
      byte[@sdbyte] := value
    else
      byte[@sdbyte+1] := value
      \sdc.fileSeek(position)
      \sdc.writeData(@sdByte+1, 1)     

  if offsetpos>0
    if (position+offsetpos)==(size-1)
      byte[@sdbyte] := backgound
    else
      byte[@sdbyte+1] := backgound
      \sdc.fileSeek(position+offsetpos)
      \sdc.writeData(@sdByte+1, 1)     

  \sdc.fileSeek(size-1)
  \sdc.writeData(@sdByte, 1)     
  \sdc.closeFile
                  
DAT '                                         UTILITY FUNCTIONS
PUB wait(sec)
  waitcnt(clkfreq*sec+cnt)

PUB waitf(div)
  waitcnt(clkfreq/div+cnt)
   
DAT '                            Debug Function(s)
PUB checkForRTCModule | s
  rtcFound := sdc.readClock
  ifnot rtcFound
    repeat s from 0 to 3
      write(s, 9, 3, YB, string(" CLOCK MODULE NOT FOUND! "))
      write(s, 10, 5, WK, string("Power down and ensure"))
      write(s, 10, 6, WK, string("that a clock modulue"))            
      write(s, 10, 7, WK, string("is present on the PCB."))
      write(s, 10, 8, WK, string("Then restart console."))
      write(s, 10, 12, GK, string("Loading Game Menu..."))
      
    waitcnt(clkfreq*10 + cnt)  
    \sdc.bootPartition(string("/GAMEMENU/GAMEMENU.BIN"))
                        
    repeat 'endless loop if game menu doesn't load off SD card

DAT 'vconfig
vconfig long    0 << 9 | %%103_3     'with sync   %010011_11  
        long    1 << 9 | %%103_0     'no sync     %010011_00      
        long    2 << 9 | %%103_0     '   
        long    3 << 9 | %%103_0     '           
sdByte  long 0
DAT

{
  *** This technique is from Roger Williams' vga_40x18_rom_demo.spin v1.1 ***
  
  The user defined chars must be reversed and interleaved like the ROM font,
  .... This example shows a way to draw the font forward and un-interleaved
  so that the proptool will assemble it correctly.  Draw the left/lsb=0
  character with 2's and the right/lsb=1 character with 1's.  The reverse
}

user_font 'beginning address of user's custom font data

              'Cust Char 0: FIT-Blank     Cust Char 1: Vert-Split
C00_01  long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32  'Note: the slightly shortend underscore makes 
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32  '...them more countable for filling in blanks

              'Cust Char 2: Banner        Cust Char 3: 3px Top Bar
C02_03  long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32                                                         
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32

              'Cust Char 4: Left V-Bar    Cust Char 5: Right V-Bar        
C04_05  long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32                                                         
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32                 

              'Cust Char 6: Diag1         Cust Char 7: Diag2                in the order of an 'o' clk-wise from up-lf corner (6-9)
C06_07  long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32  
        long  %%00000000_00000002 >< 32 + %%10000000_00000000 >< 32
        long  %%00000000_00000002 >< 32 + %%10000000_00000000 >< 32
        long  %%00000000_00000022 >< 32 + %%11000000_00000000 >< 32
        long  %%00000000_00000022 >< 32 + %%11000000_00000000 >< 32
        long  %%00000000_00000222 >< 32 + %%11100000_00000000 >< 32
        long  %%00000000_00000222 >< 32 + %%11100000_00000000 >< 32                                                         
        long  %%00000000_00002222 >< 32 + %%11110000_00000000 >< 32
        long  %%00000000_00002222 >< 32 + %%11110000_00000000 >< 32
        long  %%00000000_00022222 >< 32 + %%11111000_00000000 >< 32
        long  %%00000000_00022222 >< 32 + %%11111000_00000000 >< 32
        long  %%00000000_00222222 >< 32 + %%11111100_00000000 >< 32
        long  %%00000000_00222222 >< 32 + %%11111100_00000000 >< 32
        long  %%00000000_02222222 >< 32 + %%11111110_00000000 >< 32
        long  %%00000000_02222222 >< 32 + %%11111110_00000000 >< 32
        long  %%00000000_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%00000002_22222222 >< 32 + %%11111111_10000000 >< 32
        long  %%00000002_22222222 >< 32 + %%11111111_10000000 >< 32
        long  %%00000022_22222222 >< 32 + %%11111111_11000000 >< 32
        long  %%00000022_22222222 >< 32 + %%11111111_11000000 >< 32
        long  %%00000222_22222222 >< 32 + %%11111111_11100000 >< 32
        long  %%00000222_22222222 >< 32 + %%11111111_11100000 >< 32
        long  %%00002222_22222222 >< 32 + %%11111111_11110000 >< 32
        long  %%00002222_22222222 >< 32 + %%11111111_11110000 >< 32
        long  %%00022222_22222222 >< 32 + %%11111111_11111000 >< 32
        long  %%00022222_22222222 >< 32 + %%11111111_11111000 >< 32
        long  %%00222222_22222222 >< 32 + %%11111111_11111100 >< 32
        long  %%00222222_22222222 >< 32 + %%11111111_11111100 >< 32
        long  %%02222222_22222222 >< 32 + %%11111111_11111110 >< 32
        long  %%02222222_22222222 >< 32 + %%11111111_11111110 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32          

              'Cust Char 8: Diag3         Cust Char 9: Diag4         
C08_09  long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32  
        long  %%22222222_22222220 >< 32 + %%01111111_11111111 >< 32
        long  %%22222222_22222220 >< 32 + %%01111111_11111111 >< 32
        long  %%22222222_22222200 >< 32 + %%00111111_11111111 >< 32
        long  %%22222222_22222200 >< 32 + %%00111111_11111111 >< 32
        long  %%22222222_22222000 >< 32 + %%00011111_11111111 >< 32
        long  %%22222222_22222000 >< 32 + %%00011111_11111111 >< 32                                                         
        long  %%22222222_22220000 >< 32 + %%00001111_11111111 >< 32
        long  %%22222222_22220000 >< 32 + %%00001111_11111111 >< 32
        long  %%22222222_22200000 >< 32 + %%00000111_11111111 >< 32
        long  %%22222222_22200000 >< 32 + %%00000111_11111111 >< 32
        long  %%22222222_22000000 >< 32 + %%00000011_11111111 >< 32
        long  %%22222222_22000000 >< 32 + %%00000011_11111111 >< 32
        long  %%22222222_20000000 >< 32 + %%00000001_11111111 >< 32
        long  %%22222222_20000000 >< 32 + %%00000001_11111111 >< 32
        long  %%22222222_00000000 >< 32 + %%00000000_11111111 >< 32
        long  %%22222222_00000000 >< 32 + %%00000000_11111111 >< 32
        long  %%22222220_00000000 >< 32 + %%00000000_01111111 >< 32
        long  %%22222220_00000000 >< 32 + %%00000000_01111111 >< 32
        long  %%22222200_00000000 >< 32 + %%00000000_00111111 >< 32
        long  %%22222200_00000000 >< 32 + %%00000000_00111111 >< 32
        long  %%22222000_00000000 >< 32 + %%00000000_00011111 >< 32
        long  %%22222000_00000000 >< 32 + %%00000000_00011111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22200000_00000000 >< 32 + %%00000000_00000111 >< 32
        long  %%22200000_00000000 >< 32 + %%00000000_00000111 >< 32
        long  %%22000000_00000000 >< 32 + %%00000000_00000011 >< 32
        long  %%22000000_00000000 >< 32 + %%00000000_00000011 >< 32
        long  %%20000000_00000000 >< 32 + %%00000000_00000001 >< 32
        long  %%20000000_00000000 >< 32 + %%00000000_00000001 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32 

              'Cust Char 10: bullet       Cust Char 11: Reg Blank         
C10_11  long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32  
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002200_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002220_00000000 >< 32 + %%00000000_00000000 >< 32                                                         
        long  %%00002222_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_20000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22200000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22220000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222200 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222220 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222220 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222200 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22220000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22200000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_20000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_00000000 >< 32 + %%01111111_11111100 >< 32
        long  %%00002220_00000000 >< 32 + %%01111111_11111100 >< 32
        long  %%00002200_00000000 >< 32 + %%01111111_11111100 >< 32
        long  %%00002000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32

              'Cust Char 12: Block         Cust Char 13: Low blank         
C12_13  long  %%22222222_22222222 >< 32 + %%01111111_11111100 >< 32  
        long  %%22222222_22222222 >< 32 + %%01111111_11111100 >< 32
        long  %%22222222_22222222 >< 32 + %%01111111_11111100 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32                                                         
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32 

              'Cust Char 14: Block         Cust Char 15: Mesh         
C14_15  long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32  
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32                                                         
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32 

              'Cust Char 16: TBD          Cust Char 17: ellipsis...         
C16_17  long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32  
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000022 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000022 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000022 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000222 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000220 >< 32 + %%00000000_00000000 >< 32                                                         
        long  %%00000000_00002220 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00002220 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00002220 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00022200 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00022200 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00022200 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00222000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00222000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_02222000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_02220000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_22220000 >< 32 + %%00000000_00000000 >< 32
        long  %%20000000_22220000 >< 32 + %%00000000_00000000 >< 32
        long  %%22000002_22200000 >< 32 + %%00000000_00000000 >< 32
        long  %%02200002_22200000 >< 32 + %%00000000_00000000 >< 32
        long  %%02220022_22200000 >< 32 + %%00000000_00000000 >< 32
        long  %%00222222_22000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00222222_22000000 >< 32 + %%00111001_11001110 >< 32
        long  %%00022222_22000000 >< 32 + %%00111001_11001110 >< 32
        long  %%00022222_20000000 >< 32 + %%00111001_11001110 >< 32
        long  %%00002222_20000000 >< 32 + %%00111001_11001110 >< 32
        long  %%00002222_20000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000222_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000222_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32

              'Cust Char X: Logo-TBar     Cust Char X: Logo-BBar   
C18_19  long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32                                                         
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32        

DAT
{{

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                  TERMS OF USE: MIT License
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
// Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}             